#include "PmUsr.h"
#include <math.h>

#define ABS(a) ((a)<0.0e0?-(a):(a))

extern void Progress(double);

long PmDecompose(PmDimSystem,PmStatus,PmProfile,PmAa,PmError)
/*----------------------------------------------------------------------*/
/*  PROFILE MATRIX SERVICE :  Decompose Symmetric Profile Matrix        */
/*                            A = L * D * L(**t)                        */
/*----------------------------------------------------------------------*/
    long PmDimSystem ;          /* number of rows/columns in matrix A   */
    long PmStatus[]  ;          /* status of the rows of matrix A :     */
                                /* CPmPrimal   / CPmDual     /          */
                                /* CPmNegative / CPmInactive /          */
    long PmProfile[] ;          /* profile of the rows of matrix A      */
    double *PmAa[]   ;          /* pointer to virtual start of row of A */
    long *PmError    ;          /* error line number                    */
/*----------------------------------------------------------------------*/
/*   The given matrix A is replaced by the left-triangular matrix L     */
/*   with positive diagonal elements. The diagonal matrix D contains    */
/*   elements +1 or -1. The lines containing -1 are noted in the vector */
/*   PmStatus[]. The vectors PmProfile[] and PmAa[] remain unchanged.   */
/*                                                                      */
/*   VALUES OF THE FUNCTION                                             */
/*                                                                      */
/*   CPmOK      matrix A has been decomposed                            */
/*   CPmPivot   diagonal element Zero in line PmAa[*PmError]            */
/*                                                                      */
/*   STATUS OF MATRIX A                                                 */
/*                                                                      */
/*   The vector PmStatus[] characterizes the individual rows of matrix  */
/*   A as well as the associated variables of the system of equations.  */
/*   The associated variables of row i of the matrix are the variable   */
/*   PmXx[i] with which the diagonal element of the row is multiplied   */
/*   and the variable bb[i]. The following states are defined:          */
/*                                                                      */
/*   PmStatus[i] = CPmInactive :  row is inactive                       */
/*   Row and column i of matrix A as well as the variables PmXx[i] and  */
/*   PmBb[i] are ignored in the solution procedure.                     */
/*                                                                      */
/*   PmStatus[i] = CPmPrimal   :  primal variable PmXx[i] is specified  */
/*   The dual variable PmBb[i] is determined in the solution procedure. */
/*                                                                      */
/*   PmStatus[i] = CPmDual     :  dual variable PmBb[i] is specified    */
/*   The primal variable PmXx[i] is determined in solution procedure.   */
/*                                                                      */
/*   PmStatus[i] = CPmNegative :  diagonal element D(i,i) is negative   */
/*   This value is set during the decomposition of the matrix A to      */
/*   register a negative value -1 in the diagonal matrix D. The status  */
/*   CPmNegative can only be given to lines with the initial status     */
/*   CPmDual.                                                           */
/*                                                                      */
/*                                                                      */
/*   PROFILE OF MATRIX A                                                */
/*                                                                      */
/*   The vector PmProfile[i] contains the Index m of the first element  */
/*   A(i,m) in Line i which differs from Zero.                          */
/*                                                                      */
/*                                                                      */
/*   COEFFICIENTS OF MATRIX A                                           */
/*                                                                      */
/*   The coefficients of the left triangle of matrix A including the    */
/*   diagonal elements are stored row-wise. The  function PmDecompose() */
/*   adresses the coefficients of matrix A with the pointer vector      */
/*   PmAa[] described below. All elements within the profile (including */
/*   Zero-elements) must be initialized. The function PmDecompose()     */
/*   modifies these coefficients as follows:                            */
/*                                                                      */
/*   - Inactive rows and columns with the same index are not changed.   */
/*   - Rows and columns associated with a specified primal variable     */
/*     PmXx[i] are not modified.                                        */
/*   - Rows and columns associated with a specified dual variable       */
/*     PmBb[i] are decomposed. The coefficients of the resulting left   */
/*     triangular matrix A replace the given coefficients. On the       */
/*     diagonal, the inverse of the diagonal element is stored.         */
/*   - Positive elements +1 of diagonal matrix D of the decomposition   */
/*     are denoted by retaining the status CPmDual. Negative values -1  */
/*     are registered by changing the status of the row to CPmNegative. */
/*                                                                      */
/*                                                                      */
/*   POINTERS TO THE LINES OF MATRIX A                                  */
/*                                                                      */
/*   The pointer vector PmAa[i] contains for each row of matrix A the   */
/*   virtual address of its first element A(i,1), even if this element  */
/*   falls outside the profile of the matrix and is not actually stored.*/
/*   The virtual address is set so that it leads to the correct address */
/*   for the stored diagonal element A(i,i).                            */
/*                                                                      */
/*                                                                      */
/*   COMPUTATIONAL ALGORITHM                                            */
/*                                                                      */
/*   If all rows of matrix A have the status CPmDual, the decomposition */
/*   is performed in the followig steps:                                */
/*                                                                      */
/*   Loop on the rows    :  row     = 1,...,PmDimSystem                 */
/*   Loop on the columns :  column  = PmProfile[row],...,row            */
/*   Critical profile    :  p= Max(PmProfile[row],PmProfile[col])       */
/*   Decomposed element  :  A(i,m) = A(i,m) - (Sum from p to            */
/*                          col-1){A(row,j)*A(col,j)}                   */
/*   Case row != column  :  A(row,col) = A(line,col)/A(col,col)         */
/*   Case row == column  :  A(row,row) = sqrt(A(row,row))               */
/*                                                                      */
/*   If a row is in status CPmInactive or CPmPrimal, all operations on  */
/*   this row and on the column with the same index are suppressed.     */
/*                                                                      */
/*   If the diagonal element A(i,i) becomes negative, the square root   */
/*   is taken for the absolute value and the status of the row is set   */
/*   to CPmNegative.                                                    */
/*                                                                      */
/*   If the diagonal element A(i,i) becomes Zero within the given limit */
/*   of accuracy, the decomposition is broken off and the error code    */
/*   CPmError is returned.                                              */
/*----------------------------------------------------------------------*/

{   long SysRow       ;         /* number of the current row            */
    long SysColumn    ;         /* number of the current column         */
    long loop         ;         /* loop variable for the scalar product */
    long start        ;         /* start index of overlap row /column   */
    double diagonal   ;         /* diagonal element of matrix           */
    double eps        ;         /* limit for Zero pivot element         */
    double *PtrRow    ;         /* pointer virtual start line           */
    double *PtrColumn ;         /* pointer virtual start column         */

    double n;

/*...Compute limit on Zero value for pivot element......................*/

    eps = 0.0e0;
    for(SysRow=0; SysRow<PmDimSystem; SysRow++)
	eps += ABS(PmAa[SysRow][SysRow]);
    eps *= 1.0e-12/PmDimSystem ;

/*...DECOMPOSE BY ROWS..................................................*/

/*...Loop on all rows of matrix A with specified dual variable..........*/

    n = (double) PmDimSystem;

    for(SysRow=0; SysRow < PmDimSystem; SysRow++)
    {

      Progress( (double) SysRow/n );

   if (PmStatus[SysRow] != CPmDual) continue ;
        PtrRow = PmAa[SysRow];

/*...Loop on all columns of matrix A with specified dual variable.......*/

        for(SysColumn=PmProfile[SysRow]; SysColumn<SysRow; SysColumn++)
        {   if ((PmStatus[SysColumn] == CPmDual)||
                                    (PmStatus[SysColumn] == CPmNegative))
            {   PtrColumn = PmAa[SysColumn];

/*...Scalar product of overlapping parts of row and column :............*/
/*   Lines of L**t with status CPmDual enter into scalar product with   */
/*   positive sign, rows with status CPmNegative with negative sign     */

                if (PmProfile[SysColumn] >(start=PmProfile[SysRow]))
                    start = PmProfile[SysColumn] ;

                for(loop=start; loop < SysColumn; loop++)
                {   if (PmStatus[loop] == CPmDual)
                        PtrRow[SysColumn] -= PtrRow[loop]*PtrColumn[loop];
                    else if (PmStatus[loop] == CPmNegative)
                        PtrRow[SysColumn] += PtrRow[loop]*PtrColumn[loop];
                }
	    }
/*...Divide with the diagonal element of the column and reduce the......*/
/*   the diagonal element of the row                                    */

	    if (PmStatus[SysColumn] == CPmDual)
	    {   PtrRow[SysColumn] *= PtrColumn[SysColumn] ;
		PtrRow[SysRow] -= PtrRow[SysColumn]*PtrRow[SysColumn];
	    }
	    else if (PmStatus[SysColumn] == CPmNegative)
	    {   PtrRow[SysColumn] *= (-PtrColumn[SysColumn]) ;
		PtrRow[SysRow] += PtrRow[SysColumn]*PtrRow[SysColumn];
	    }
	}

/*...Test the diagonal element of the row for Zero......................*/

	if (((diagonal = PtrRow[SysRow]) < eps) && (diagonal > (-eps)))
	{   *PmError = SysRow ;
	    return (CPmPivot) ;
	}

/*...Note Negative diagonal element.....................................*/

	if (diagonal < 0.0e0)
	{   PmStatus[SysRow] =  CPmNegative    ;
	    PtrRow[SysRow]   = -PtrRow[SysRow] ;
	}

/*...Store inverse value of diagonal element............................*/

	PtrRow[SysRow] = 1.0 / sqrt(PtrRow[SysRow]) ;
    }

    return (CPmOK) ;
}
