/*
  Datei : PmSolveEqn.c
*/

#include "PmUsr.h"


void PmSolveEqn(PmDimSystem,PmStatus,PmProfile,PmAa,PmXx,PmBb)
/*----------------------------------------------------------------------*/
/*  PROFILE MATRIX SERVICE :  Solve the Linear Equations : A * x = b    */
/*----------------------------------------------------------------------*/
    long PmDimSystem ;          /* number of rows/columns in matrix A   */
    long PmStatus[]  ;          /* status of the rows of matrix A :     */
                                /* CPmPrimal   / CPmDual     /          */
                                /* CPmNegative / CPmInactive /          */
    long PmProfile[] ;          /* profile of the rows of matrix A      */
    double *PmAa[]   ;          /* Pointer to virtual start of row of A */
    double *PmXx     ;          /* Pointer to primal vector X           */
    double *PmBb     ;          /* Pointer to dual   vector B           */
/*----------------------------------------------------------------------*/
/*  The matrix A is given in the decomposed form L*D*L(**t) described   */
/*  in function PmDecompose(). In this form, the lines of A associated  */
/*  with specified values PmXx[i] contain the coefficients of A before  */
/*  decomposition of the matrix. The vector PmProfile[] describing the  */
/*  structure of A is also described in PmDecompose().                  */
/*                                                                      */
/*  For each row of A, the vector PmStatus[i] specifies which of the    */
/*  variables PmXx[i] or PmBb[i] are specified:                         */
/*                                                                      */
/*  PmStatus[i] = CPmInactive  :  row and column i are inactive         */
/*                CPmPrimal    :  value of PmXx[i] is specified         */
/*                CPmDual      :  value of PmBb[i] is specified         */
/*                                D[i][i] > 0                           */
/*                CPmNegative  :  value of PmBb[i] is specified         */
/*                                D[i][i] < 0                           */
/*                                                                      */
/*  In the vector PmXx[], only the specified values in the rows with    */
/*  status CPmPrimal must be set when PmSolve() is called. In vector    */
/*  PmBb[], all rows must be set. The rows with state CPmDual or        */
/*  CPmNegative must contain the specified value. The rows with status  */
/*  CPpPrimal must be set to zero or another desired value (eg. the     */
/*  loads applied at the supports of a structure). This value is used   */
/*  in the determination of the variable PmBb[i] (eg. the reaction at   */
/*  the support).                                                       */
/*                                                                      */
/*  The decomposed matrix A and the vectors PmStatus and PmProfile are  */
/*  not modified in PmSolve(). The specified values in the vectors      */
/*  PmXx[] and PmBb[] are preserved. The unknown variables in PmXx[]    */
/*  and PmBb[] are computed and stored in these vectors.                */
/*----------------------------------------------------------------------*/

{    long SysRow       ;        /* row number in system matrix A        */
     long SysColumn    ;        /* column number in system matrix A     */
     double *PtrRow    ;        /* pointer to fictive start of row      */
     double *PtrColumn ;        /* pointer to fictive start of column   */

/*...Copy specified dual variables into vector X to preserve B..........*/

     for(SysRow=0; SysRow < PmDimSystem; SysRow++)
         if ((PmStatus[SysRow] == CPmDual) ||
                                        (PmStatus[SysRow] == CPmNegative))
             PmXx[SysRow] = PmBb[SysRow] ;

/*...Substitute the specified primal variables above the diagonal.......*/

/*...Loop on all columns with specified associated X-value..............*/

     for (SysColumn=0; SysColumn < PmDimSystem; SysColumn++)
         if (PmStatus[SysColumn] == CPmPrimal)
         {   PtrColumn = PmAa[SysColumn];

/*...Substitute X-value in column on all rows with specified B-value....*/

             for(SysRow=PmProfile[SysColumn]; SysRow<SysColumn; SysRow++)
             {   if ((PmStatus[SysRow] == CPmDual) ||
                                       (PmStatus[SysRow] == CPmNegative))
                     PmXx[SysRow] -= PtrColumn[SysRow]*PmXx[SysColumn];
             }
         }

/*...Substitute the specified primal variables below the diagonal.......*/

/*...Loop on all rows with a specified B-value..........................*/

     for(SysRow=0; SysRow<PmDimSystem; SysRow++)
     {   if ((PmStatus[SysRow] == CPmDual) ||
                                       (PmStatus[SysRow] == CPmNegative))
         {   PtrRow = PmAa[SysRow] ;

/*...Substitute all specified X-values in the row.......................*/

             for(SysColumn=PmProfile[SysRow];SysColumn<SysRow;SysColumn++)
             {   if (PmStatus[SysColumn] == CPmPrimal)
                     PmXx[SysRow] -= PtrRow[SysColumn]*PmXx[SysColumn] ;
             }
         }
     }

/*...FORWARD SWEEP TO COMPUTE PRIMAL VARIABLES..........................*/

/*...Loop on all rows with specified B-values...........................*/

     for(SysRow=0; SysRow < PmDimSystem; SysRow++)
     {   if ((PmStatus[SysRow] == CPmDual) ||
                                       (PmStatus[SysRow] == CPmNegative))
         {   PtrRow = PmAa[SysRow];

/*...Substitute current X-values left of the diagonal...................*/

             for(SysColumn=PmProfile[SysRow];SysColumn<SysRow;SysColumn++)
             {   if ((PmStatus[SysColumn] == CPmDual) ||
                                     (PmStatus[SysColumn] == CPmNegative))
                     PmXx[SysRow] -= PtrRow[SysColumn]*PmXx[SysColumn] ;
             }

/*...Multiply with the stored inverse of the diagonal element...........*/

             PmXx[SysRow] *= PtrRow[SysRow] ;
         }
     }

/*...Diagonal matrix with elements +1 or -1.............................*/

     for(SysRow=0; SysRow<PmDimSystem; SysRow++)
     {   if (PmStatus[SysRow]== CPmNegative)
             PmXx[SysRow] = -PmXx[SysRow];
     }

/*...BACKWARD SWEEP TO COMPUTE PRIMAL VARIABLES.........................*/

/*...Loop on all columns with computed associated X-value...............*/

     for(SysColumn=PmDimSystem-1; SysColumn>=0; SysColumn--)
     {   if ((PmStatus[SysColumn] == CPmDual) ||
                                   (PmStatus[SysColumn] == CPmNegative))
         {   PtrColumn = PmAa[SysColumn] ;

/*...Multiply with the stored inverse of the diagonal element...........*/

             PmXx[SysColumn] *= PtrColumn[SysColumn];

/*...Substitute primal variables in all rows with specified dual var....*/

             for (SysRow=PmProfile[SysColumn];SysRow<SysColumn;SysRow++)
             {   if ((PmStatus[SysRow] == CPmDual) ||
                                      (PmStatus[SysRow] == CPmNegative))
                     PmXx[SysRow] -= PtrColumn[SysRow]*PmXx[SysColumn] ;
             }
         }
     }

/*...COMPUTE DUAL VARIABLES.............................................*/

     for(SysColumn=0; SysColumn<PmDimSystem; SysColumn++)
	 if (PmStatus[SysColumn] == CPmPrimal)
	     PmBb[SysColumn] = -PmBb[SysColumn] ;

/*...Substitute primal variables above and on the diagonal..............*/

     for(SysColumn=0; SysColumn<PmDimSystem; SysColumn++)
     {   PtrColumn = PmAa[SysColumn] ;

/*...Loop on all rows with specified primal variables...................*/

         for(SysRow = PmProfile[SysColumn]; SysRow<=SysColumn; SysRow++)
         {   if (PmStatus[SysRow] == CPmPrimal)
                 PmBb[SysRow] += PtrColumn[SysRow]*PmXx[SysColumn];
         }
     }

/*...Substitute primal variables below the diagonal.....................*/

/*...Loop on all rows with specified primal variables...................*/

     for(SysRow=0; SysRow < PmDimSystem; SysRow++)
     {   if (PmStatus[SysRow] == CPmPrimal)
         {   PtrRow = PmAa[SysRow];

/*...Loop on all columns................................................*/

             for(SysColumn = PmProfile[SysRow]; SysColumn < SysRow;
                                                             SysColumn++)
             {   PmBb[SysRow] += PtrRow[SysColumn]*PmXx[SysColumn];
             }
         }
     }
     return ;
}
